from torchmeta.datasets.cifar100 import FC100
from torchmeta.datasets.miniimagenet import MiniImagenet
from torchmeta.datasets.omniglot import Omniglot
from torchmeta.datasets.tieredimagenet import TieredImagenet
from torchmeta.datasets.tcga import TCGA

from torchmeta.datasets import helpers

__all__ = ['TCGA', 'Omniglot', 'MiniImagenet', 'TieredImagenet', 'FC100', 'helpers']
